__all__ = [
    'plot_or_save', 
    'plot_stack',
    'plot_stack_bar',
    'plot_boxplot_cat_num_var',
    'plot_correlation_matrix',
    'plot_gauge',
    'plot_bar',
    'save_plot',
    'save_dir',
    'verbose'
    ]

from transparentai.plots.plots import *